/* Compute MD5 of strings (see MD5 procedure below)
   Author: Frank Ellermann. 1 March 2004
*/

/* MD5 test strings  */

   signal on novalue

   X = ''
   Y = 'd41d8cd98f00b204e9800998ecf8427e'
   if MD5( X ) = Y then say 'ok.: empty'

   X = 'a'
   Y = '0cc175b9c0f1b6a831c399e269772661'
   if MD5( X ) = Y then say 'ok.:' X

   X = 'abc'
   Y = '900150983cd24fb0d6963f7d28e17f72'
   if MD5( X ) = Y then say 'ok.:' X

   X = 'message digest'
   Y = 'f96b697d7cb7938d525a2f31aaf161d0'
   if MD5( X ) = Y then say 'ok.:' X

   X = 'abcdefghijklmnopqrstuvwxyz'
   Y = 'c3fcd3d76192e4007dfb496cca67e13b'
   if MD5( X ) = Y then say 'ok.:' X

   X = 'ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz'
   X = X || '0123456789'
  Y = 'd174ab98d277d9f5a5611c2c9f419d9f'
   if MD5( X ) = Y then say 'ok.: ABC ... 89'

   X = copies( '1234567890', 8 )
   Y = '57edf4a22be3c955ac49da2e2107b67a'
   if MD5( X ) = Y then say 'ok.: 123....123...'

   X = '<1896.697170952@dbc.mtview.ca.us>tanstaaf'
   Y = 'c4c9334bac560ecc979e58001b3e22fb'
   if MD5( X ) = Y then say 'ok.: RfC 1939 APOP'
   exit

/* -------------------------------------------------------------- */
/* MD5 procedures derived from Daniel Hellerstein's REXX_MD5.ZIP, */
/* see <URL:http://hobbes.nmsu.edu/pub/os2/dev/rexx/rexx_md5.zip> */

MD5   :  procedure               /* for MD5 details see RfC 1321  */
   numeric digits 11             /* 32 bit addition + carry digit */
   parse arg MSG                 /* if MSG < 12,500,000,000 bytes */

   LEN = 8 * length( MSG )       /* MD5 operates on bit messages: */
   PAD = LEN // 512              /* pad to length 448 modulo 512  */
   if PAD < 448   then  PAD = 448 - PAD
                  else  PAD = 960 - PAD
   PAD = left( '80'x, PAD / 8, '00'x )

   MSG = MSG || PAD || reverse( x2c( d2x( LEN, 16 )))

   A = '67452301'x      ;  B = 'EFCDAB89'x
   C = '98BADCFE'x      ;  D = '10325476'x

   S11 = 7  ;  S12 = 12 ;  S13 = 17 ;  S14 = 22
   S21 = 5  ;  S22 = 9  ;  S23 = 14 ;  S24 = 20
   S31 = 4  ;  S32 = 11 ;  S33 = 16 ;  S34 = 23
   S41 = 6  ;  S42 = 10 ;  S43 = 15 ;  S44 = 21

   do N = 1 to ( length( MSG ) / 4 ) / 16
      K = N * 64 - 63
      do J = 0 to 15
         M.J = c2d( reverse( substr( MSG, K + J * 4, 4 )))
      end J

      AA = A   ;  BB = B   ;  CC = C   ;  DD = D

      A = MD5.1( A, B, C, D, S11, M.0 + 3614090360 )  /*  1 */
      D = MD5.1( D, A, B, C, S12, M.1 + 3905402710 )  /*  2 */
      C = MD5.1( C, D, A, B, S13, M.2 +  606105819 )  /*  3 */
      B = MD5.1( B, C, D, A, S14, M.3 + 3250441966 )  /*  4 */
      A = MD5.1( A, B, C, D, S11, M.4 + 4118548399 )  /*  5 */
      D = MD5.1( D, A, B, C, S12, M.5 + 1200080426 )  /*  6 */
      C = MD5.1( C, D, A, B, S13, M.6 + 2821735955 )  /*  7 */
      B = MD5.1( B, C, D, A, S14, M.7 + 4249261313 )  /*  8 */
      A = MD5.1( A, B, C, D, S11, M.8 + 1770035416 )  /*  9 */
      D = MD5.1( D, A, B, C, S12, M.9 + 2336552879 )  /* 10 */
      C = MD5.1( C, D, A, B, S13, M.10+ 4294925233 )  /* 11 */
      B = MD5.1( B, C, D, A, S14, M.11+ 2304563134 )  /* 12 */
      A = MD5.1( A, B, C, D, S11, M.12+ 1804603682 )  /* 13 */
      D = MD5.1( D, A, B, C, S12, M.13+ 4254626195 )  /* 14 */
      C = MD5.1( C, D, A, B, S13, M.14+ 2792965006 )  /* 15 */
      B = MD5.1( B, C, D, A, S14, M.15+ 1236535329 )  /* 16 */

      A = MD5.2( A, B, C, D, S21, M.1 + 4129170786 )  /* 17 */
      D = MD5.2( D, A, B, C, S22, M.6 + 3225465664 )  /* 18 */
      C = MD5.2( C, D, A, B, S23, M.11+  643717713 )  /* 19 */
      B = MD5.2( B, C, D, A, S24, M.0 + 3921069994 )  /* 20 */
      A = MD5.2( A, B, C, D, S21, M.5 + 3593408605 )  /* 21 */
      D = MD5.2( D, A, B, C, S22, M.10+   38016083 )  /* 22 */
      C = MD5.2( C, D, A, B, S23, M.15+ 3634488961 )  /* 23 */
      B = MD5.2( B, C, D, A, S24, M.4 + 3889429448 )  /* 24 */
      A = MD5.2( A, B, C, D, S21, M.9 +  568446438 )  /* 25 */
      D = MD5.2( D, A, B, C, S22, M.14+ 3275163606 )  /* 26 */
      C = MD5.2( C, D, A, B, S23, M.3 + 4107603335 )  /* 27 */
      B = MD5.2( B, C, D, A, S24, M.8 + 1163531501 )  /* 28 */
      A = MD5.2( A, B, C, D, S21, M.13+ 2850285829 )  /* 29 */
      D = MD5.2( D, A, B, C, S22, M.2 + 4243563512 )  /* 30 */
      C = MD5.2( C, D, A, B, S23, M.7 + 1735328473 )  /* 31 */
      B = MD5.2( B, C, D, A, S24, M.12+ 2368359562 )  /* 32 */

      A = MD5.3( A, B, C, D, S31, M.5 + 4294588738 )  /* 33 */
      D = MD5.3( D, A, B, C, S32, M.8 + 2272392833 )  /* 34 */
      C = MD5.3( C, D, A, B, S33, M.11+ 1839030562 )  /* 35 */
      B = MD5.3( B, C, D, A, S34, M.14+ 4259657740 )  /* 36 */
      A = MD5.3( A, B, C, D, S31, M.1 + 2763975236 )  /* 37 */
      D = MD5.3( D, A, B, C, S32, M.4 + 1272893353 )  /* 38 */
      C = MD5.3( C, D, A, B, S33, M.7 + 4139469664 )  /* 39 */
      B = MD5.3( B, C, D, A, S34, M.10+ 3200236656 )  /* 40 */
      A = MD5.3( A, B, C, D, S31, M.13+  681279174 )  /* 41 */
      D = MD5.3( D, A, B, C, S32, M.0 + 3936430074 )  /* 42 */
      C = MD5.3( C, D, A, B, S33, M.3 + 3572445317 )  /* 43 */
      B = MD5.3( B, C, D, A, S34, M.6 +   76029189 )  /* 44 */
      A = MD5.3( A, B, C, D, S31, M.9 + 3654602809 )  /* 45 */
      D = MD5.3( D, A, B, C, S32, M.12+ 3873151461 )  /* 46 */
      C = MD5.3( C, D, A, B, S33, M.15+  530742520 )  /* 47 */
      B = MD5.3( B, C, D, A, S34, M.2 + 3299628645 )  /* 48 */

      A = MD5.4( A, B, C, D, S41, M.0 + 4096336452 )  /* 49 */
      D = MD5.4( D, A, B, C, S42, M.7 + 1126891415 )  /* 50 */
      C = MD5.4( C, D, A, B, S43, M.14+ 2878612391 )  /* 51 */
      B = MD5.4( B, C, D, A, S44, M.5 + 4237533241 )  /* 52 */
      A = MD5.4( A, B, C, D, S41, M.12+ 1700485571 )  /* 53 */
      D = MD5.4( D, A, B, C, S42, M.3 + 2399980690 )  /* 54 */
      C = MD5.4( C, D, A, B, S43, M.10+ 4293915773 )  /* 55 */
      B = MD5.4( B, C, D, A, S44, M.1 + 2240044497 )  /* 56 */
      A = MD5.4( A, B, C, D, S41, M.8 + 1873313359 )  /* 57 */
      D = MD5.4( D, A, B, C, S42, M.15+ 4264355552 )  /* 58 */
      C = MD5.4( C, D, A, B, S43, M.6 + 2734768916 )  /* 59 */
      B = MD5.4( B, C, D, A, S44, M.13+ 1309151649 )  /* 60 */
      A = MD5.4( A, B, C, D, S41, M.4 + 4149444226 )  /* 61 */
      D = MD5.4( D, A, B, C, S42, M.11+ 3174756917 )  /* 62 */
      C = MD5.4( C, D, A, B, S43, M.2 +  718787259 )  /* 63 */
      B = MD5.4( B, C, D, A, S44, M.9 + 3951481745 )  /* 64 */

      A = x2c( d2x( c2d( AA ) + c2d( A ), 8 ))
      B = x2c( d2x( c2d( BB ) + c2d( B ), 8 ))
      C = x2c( d2x( c2d( CC ) + c2d( C ), 8 ))
      D = x2c( d2x( c2d( DD ) + c2d( D ), 8 ))
   end N

   A = reverse( A )  ;  B = reverse( B )
   C = reverse( C )  ;  D = reverse( D )
   return translate( c2x( A || B || C || D ), 'abcdef', 'ABCDEF' )

MD5.1:   procedure               /* function used in MD5 round 1: */
   parse arg A, B, C, D, S, M
   C = bitor( bitand( B, C ), bitand( D, bitxor( B,, 'FF'x )))
   signal MD5..

MD5.2:   procedure               /* function used in MD5 round 2: */
   parse arg A, B, C, D, S, M
   C = bitor( bitand( B, D ), bitand( C, bitxor( D,, 'FF'x )))
   signal MD5..

MD5.3:   procedure               /* function used in MD5 round 3: */
   parse arg A, B, C, D, S, M
   C = bitxor( B, bitxor( C, D ))
   signal MD5..

MD5.4:   procedure               /* function used in MD5 round 4: */
   parse arg A, B, C, D, S, M
   C = bitxor( C, bitor( B, bitxor( D,, 'FF'x )))
MD5..:
   C = x2b( d2x( c2d( A ) + c2d( C ) + M, 8 ))
   C = x2d( b2x( right( C || left( C, S ), 32 )))
   return x2c( d2x( C + c2d( B ), 8 ))



